@@ -1,10 +1,11 @@
-use std::collections::HashMap;
 use std::collections::HashSet;
 use std::ffi::OsStr;
 use std::io::Write;
 use std::path::Path;
 
 use chrono::prelude::Local;
+use indexmap::IndexMap;
+use indexmap::IndexSet;
 use itertools::Itertools;
 
 use crate::error::FatalError;
@@ -16,135 +17,184 @@ pub(crate) fn release_workspace(args: &args::ReleaseOpt) -> Result<i32, error::F
     let ws_meta = args
         .manifest
         .metadata()
-        // When evaluating dependency ordering, we need to consider optional depednencies
+        // When evaluating dependency ordering, we need to consider optional dependencies
         .features(cargo_metadata::CargoOpt::AllFeatures)
         .exec()
         .map_err(FatalError::from)?;
-    let ws_config = {
-        let mut release_config = config::Config::default();
-
-        if !args.isolated {
-            let cfg = config::resolve_workspace_config(ws_meta.workspace_root.as_std_path())?;
-            release_config.update(&cfg);
-        }
-
-        if let Some(custom_config_path) = args.custom_config.as_ref() {
-            // when calling with -c option
-            let cfg =
-                config::resolve_custom_config(Path::new(custom_config_path))?.unwrap_or_default();
-            release_config.update(&cfg);
-        }
-
-        release_config.update(&args.config.to_config());
-        release_config
-    };
-
-    let pkg_ids = cargo::sort_workspace(&ws_meta);
-
-    let (selected_pkgs, excluded_pkgs) = args.workspace.partition_packages(&ws_meta);
-    if selected_pkgs.is_empty() {
-        log::info!("No packages selected.");
-        return Ok(0);
-    }
-
     let root = git::top_level(ws_meta.workspace_root.as_std_path())?;
-    let pkg_releases: Result<HashMap<_, _>, _> = selected_pkgs
+    let ws_config = config::load_workspace_config(args, &ws_meta)?;
+
+    let member_ids = cargo::sort_workspace(&ws_meta);
+    let pkgs: Result<IndexMap<_, _>, _> = member_ids
         .iter()
-        .filter_map(|p| PackageRelease::load(args, &root, &ws_meta, p).transpose())
+        .filter_map(|p| PackageRelease::load(args, &root, &ws_meta, &ws_meta[p]).transpose())
         .map(|p| p.map(|p| (&p.meta.id, p)))
         .collect();
-    let pkg_releases = pkg_releases?;
-    let pkg_releases: Vec<_> = pkg_ids
-        .iter()
-        .filter_map(|id| pkg_releases.get(id))
-        .collect();
+    let mut pkgs = pkgs?;
 
-    if !excluded_pkgs.is_empty() {
-        // We aren't releasing these, so keep the target version the same rather than attempting to
-        // change it which can be especially bad when an absolute version is used and the target
-        // version becomes a downgrade.
-        let mut excluded_args = args.clone();
-        excluded_args.level_or_version =
-            version::TargetVersion::Relative(version::BumpLevel::Release);
-        let excluded_pkgs: Vec<_> = excluded_pkgs
-            .iter()
-            .filter(|p| pkg_ids.contains(&&p.id))
-            .filter_map(|p| PackageRelease::load(&excluded_args, &root, &ws_meta, p).transpose())
-            .collect();
-        for pkg in excluded_pkgs {
-            let pkg = match pkg {
-                Ok(pkg) => pkg,
-                Err(err) => {
-                    log::debug!("Could not analyze skipped package: {}", err);
-                    continue;
-                }
-            };
-            let crate_name = pkg.meta.name.as_str();
-            let prev_tag_name = &pkg.prev_tag;
-            if let Some((changed, lock_changed)) = changed_since(&ws_meta, &pkg, prev_tag_name) {
-                if !changed.is_empty() {
-                    log::warn!(
-                        "Disabled by user, skipping {} which has files changed since {}: {:#?}",
-                        crate_name,
-                        prev_tag_name,
-                        changed
-                    );
-                } else if lock_changed {
-                    log::warn!(
-                        "Disabled by user, skipping {} despite lock file being changed since {}",
-                        crate_name,
-                        prev_tag_name
-                    );
-                } else {
-                    log::trace!(
-                        "Disabled by user, skipping {} (no changes since {})",
-                        crate_name,
-                        prev_tag_name
-                    );
-                }
+    let (_selected_pkgs, excluded_pkgs) = args.workspace.partition_packages(&ws_meta);
+    for excluded_pkg in excluded_pkgs {
+        if !member_ids.contains(&&excluded_pkg.id) {
+            continue;
+        }
+        let pkg = &mut pkgs[&excluded_pkg.id];
+        pkg.config.release = Some(false);
+        pkg.version = None;
+
+        let crate_name = pkg.meta.name.as_str();
+        let prev_tag_name = &pkg.prev_tag;
+        if let Some((changed, lock_changed)) = changed_since(&ws_meta, pkg, prev_tag_name) {
+            if !changed.is_empty() {
+                log::warn!(
+                    "Disabled by user, skipping {} which has files changed since {}: {:#?}",
+                    crate_name,
+                    prev_tag_name,
+                    changed
+                );
+            } else if lock_changed {
+                log::warn!(
+                    "Disabled by user, skipping {} despite lock file being changed since {}",
+                    crate_name,
+                    prev_tag_name
+                );
             } else {
-                log::debug!(
-                    "Disabled by user, skipping {} (no {} tag)",
+                log::trace!(
+                    "Disabled by user, skipping {} (no changes since {})",
                     crate_name,
                     prev_tag_name
                 );
             }
+        } else {
+            log::debug!(
+                "Disabled by user, skipping {} (no {} tag)",
+                crate_name,
+                prev_tag_name
+            );
         }
     }
 
-    release_packages(args, &ws_meta, &ws_config, pkg_releases.as_slice())
+    let mut shared_max: Option<version::Version> = None;
+    let mut shared_ids = IndexSet::new();
+    for (pkg_id, pkg) in pkgs.iter() {
+        if pkg.config.shared_version() {
+            shared_ids.insert(*pkg_id);
+            let planned = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+            if shared_max
+                .as_ref()
+                .map(|max| max.full_version < planned.full_version)
+                .unwrap_or(true)
+            {
+                shared_max = Some(planned.clone());
+            }
+        }
+    }
+    if let Some(shared_max) = shared_max {
+        for shared_id in shared_ids {
+            let shared_pkg = &mut pkgs[shared_id];
+            if shared_pkg.prev_version.bare_version != shared_max.bare_version {
+                shared_pkg.version = Some(shared_max.clone());
+            }
+        }
+    }
+
+    for pkg in pkgs.values_mut() {
+        pkg.plan()?;
+    }
+
+    let pkgs: Vec<_> = pkgs
+        .into_iter()
+        .map(|(_, pkg)| pkg)
+        .filter(|p| p.config.release())
+        .collect();
+    if pkgs.is_empty() {
+        log::info!("No packages selected.");
+        return Ok(0);
+    }
+    release_packages(args, &ws_meta, &ws_config, pkgs.as_slice())
 }
 
 fn release_packages<'m>(
     args: &args::ReleaseOpt,
     ws_meta: &cargo_metadata::Metadata,
     ws_config: &config::Config,
-    pkgs: &'m [&'m PackageRelease<'m>],
+    pkgs: &'m [PackageRelease<'m>],
 ) -> Result<i32, error::FatalError> {
     let dry_run = args.dry_run();
 
     // STEP 0: Help the user make the right decisions.
     git::git_version()?;
-    let mut dirty = false;
-    if ws_config.consolidate_commits() {
-        if git::is_dirty(ws_meta.workspace_root.as_std_path())? {
-            log::error!("Uncommitted changes detected, please commit before release.");
-            dirty = true;
+
+    if git::is_dirty(ws_meta.workspace_root.as_std_path())? {
+        log::error!("Uncommitted changes detected, please commit before release.");
+        if !dry_run {
+            return Ok(101);
         }
-    } else {
-        for pkg in pkgs {
-            let cwd = pkg.package_path;
-            if git::is_dirty(cwd)? {
-                let crate_name = pkg.meta.name.as_str();
-                log::error!(
-                    "Uncommitted changes detected for {}, please commit before release.",
-                    crate_name
-                );
-                dirty = true;
+    }
+
+    let mut tag_exists = false;
+    let mut seen_tags = HashSet::new();
+    for pkg in pkgs {
+        if let Some(tag_name) = pkg.tag.as_ref() {
+            if seen_tags.insert(tag_name) {
+                let cwd = pkg.package_root;
+                if git::tag_exists(cwd, tag_name)? {
+                    let crate_name = pkg.meta.name.as_str();
+                    log::error!("Tag `{}` already exists (for `{}`)", tag_name, crate_name);
+                    tag_exists = true;
+                }
             }
         }
     }
-    if dirty {
+    if tag_exists {
+        if !dry_run {
+            return Ok(101);
+        }
+    }
+
+    let mut downgrades_present = false;
+    for pkg in pkgs {
+        if let Some(version) = pkg.version.as_ref() {
+            if version.full_version < pkg.prev_version.full_version {
+                let crate_name = pkg.meta.name.as_str();
+                log::error!(
+                    "Cannot downgrade {} from {} to {}",
+                    crate_name,
+                    version.full_version,
+                    pkg.prev_version.full_version
+                );
+                downgrades_present = true;
+            }
+        }
+    }
+    if downgrades_present {
+        if !dry_run {
+            return Ok(101);
+        }
+    }
+
+    let mut double_publish = false;
+    for pkg in pkgs {
+        if !pkg.config.publish() {
+            continue;
+        }
+        if pkg.config.registry().is_none() {
+            // While we'll publish when there is `pkg.version.is_none()`, we'll check that case
+            // during the publish
+            if let Some(version) = pkg.version.as_ref() {
+                let index = crates_index::Index::new_cargo_default()?;
+                let crate_name = pkg.meta.name.as_str();
+                if cargo::is_published(&index, crate_name, &version.full_version_string) {
+                    log::error!(
+                        "{} {} is already published",
+                        crate_name,
+                        version.full_version_string
+                    );
+                    double_publish = true;
+                }
+            }
+        }
+    }
+    if double_publish {
         if !dry_run {
             return Ok(101);
         }
@@ -253,17 +303,22 @@ fn release_packages<'m>(
     // STEP 1: Release Confirmation
     if !dry_run && !args.no_confirm {
         let prompt = if pkgs.len() == 1 {
-            let pkg = pkgs[0];
+            let pkg = &pkgs[0];
             let crate_name = pkg.meta.name.as_str();
-            let base = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
-            format!("Release {} {}?", crate_name, base.full_version_string)
+            let version = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+            format!("Release {} {}?", crate_name, version.full_version_string)
         } else {
             let mut buffer: Vec<u8> = vec![];
             writeln!(&mut buffer, "Release").unwrap();
             for pkg in pkgs {
                 let crate_name = pkg.meta.name.as_str();
-                let base = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
-                writeln!(&mut buffer, "  {} {}", crate_name, base.full_version_string).unwrap();
+                let version = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+                writeln!(
+                    &mut buffer,
+                    "  {} {}",
+                    crate_name, version.full_version_string
+                )
+                .unwrap();
             }
             write!(&mut buffer, "?").unwrap();
             String::from_utf8(buffer).expect("Only valid UTF-8 has been written")
@@ -278,15 +333,24 @@ fn release_packages<'m>(
     // STEP 2: update current version, save and commit
     let mut shared_commit = false;
     for pkg in pkgs {
-        let cwd = pkg.package_path;
+        let cwd = pkg.package_root;
         let crate_name = pkg.meta.name.as_str();
 
         if let Some(version) = pkg.version.as_ref() {
-            let prev_version_string = pkg.prev_version.full_version_string.as_str();
-            let new_version_string = version.full_version_string.as_str();
-            log::info!("Update {} to version {}", crate_name, new_version_string);
+            let prev_version_var = pkg.prev_version.bare_version_string.as_str();
+            let prev_metadata_var = pkg.prev_version.full_version.build.as_str();
+            let version_var = version.bare_version_string.as_str();
+            let metadata_var = version.full_version.build.as_str();
+            log::info!(
+                "Update {} to version {}",
+                crate_name,
+                version.full_version_string
+            );
             if !dry_run {
-                cargo::set_package_version(pkg.manifest_path, new_version_string)?;
+                cargo::set_package_version(
+                    pkg.manifest_path,
+                    version.full_version_string.as_str(),
+                )?;
             }
             update_dependent_versions(pkg, version, dry_run)?;
             if dry_run {
@@ -298,8 +362,10 @@ fn release_packages<'m>(
             if !pkg.config.pre_release_replacements().is_empty() {
                 // try replacing text in configured files
                 let template = Template {
-                    prev_version: Some(&prev_version_string),
-                    version: Some(new_version_string),
+                    prev_version: Some(prev_version_var),
+                    prev_metadata: Some(prev_metadata_var),
+                    version: Some(version_var),
+                    metadata: Some(metadata_var),
                     crate_name: Some(crate_name),
                     date: Some(NOW.as_str()),
                     tag_name: pkg.tag.as_deref(),
@@ -320,8 +386,10 @@ fn release_packages<'m>(
                 let pre_rel_hook = pre_rel_hook.args();
                 log::debug!("Calling pre-release hook: {:?}", pre_rel_hook);
                 let envs = maplit::btreemap! {
-                    OsStr::new("PREV_VERSION") => prev_version_string.as_ref(),
-                    OsStr::new("NEW_VERSION") => new_version_string.as_ref(),
+                    OsStr::new("PREV_VERSION") => prev_version_var.as_ref(),
+                    OsStr::new("PREV_METADATA") => prev_metadata_var.as_ref(),
+                    OsStr::new("NEW_VERSION") => version_var.as_ref(),
+                    OsStr::new("NEW_METADATA") => metadata_var.as_ref(),
                     OsStr::new("DRY_RUN") => OsStr::new(if dry_run { "true" } else { "false" }),
                     OsStr::new("CRATE_NAME") => OsStr::new(crate_name),
                     OsStr::new("WORKSPACE_ROOT") => ws_meta.workspace_root.as_os_str(),
@@ -342,8 +410,10 @@ fn release_packages<'m>(
                 shared_commit = true;
             } else {
                 let template = Template {
-                    prev_version: Some(prev_version_string),
-                    version: Some(new_version_string),
+                    prev_version: Some(prev_version_var),
+                    prev_metadata: Some(prev_metadata_var),
+                    version: Some(version_var),
+                    metadata: Some(metadata_var),
                     crate_name: Some(crate_name),
                     date: Some(NOW.as_str()),
                     ..Default::default()
@@ -359,10 +429,15 @@ fn release_packages<'m>(
     }
     if shared_commit {
         let shared_commit_msg = {
+            let version_var = shared_version
+                .as_ref()
+                .map(|v| v.bare_version_string.as_str());
+            let metadata_var = shared_version
+                .as_ref()
+                .map(|v| v.full_version.build.as_str());
             let template = Template {
-                version: shared_version
-                    .as_ref()
-                    .map(|v| v.full_version_string.as_str()),
+                version: version_var,
+                metadata: metadata_var,
                 date: Some(NOW.as_str()),
                 ..Default::default()
             };
@@ -381,76 +456,105 @@ fn release_packages<'m>(
 
     // STEP 3: cargo publish
     for pkg in pkgs {
-        if pkg.config.publish() {
-            let crate_name = pkg.meta.name.as_str();
-            let base = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+        if !pkg.config.publish() {
+            continue;
+        }
 
-            log::info!("Running cargo publish on {}", crate_name);
-            // feature list to release
-            let features = &pkg.features;
-            if !cargo::publish(
+        let crate_name = pkg.meta.name.as_str();
+        if pkg.config.registry().is_none() && pkg.version.is_none() {
+            let version = &pkg.prev_version;
+            let index = crates_index::Index::new_cargo_default()?;
+            if cargo::is_published(&index, crate_name, &version.full_version_string) {
+                log::warn!("Skipping publish of {} {}, assuming we are recovering from a prior failed release", crate_name, version.full_version_string);
+                continue;
+            }
+        }
+
+        log::info!("Publishing {}", crate_name);
+
+        let verify = if !pkg.config.verify() {
+            false
+        } else if dry_run && pkgs.len() != 1 {
+            log::debug!("Skipping verification to avoid unpublished dependencies from dry-run");
+            false
+        } else {
+            true
+        };
+        // feature list to release
+        let features = &pkg.features;
+        if !cargo::publish(
+            dry_run,
+            verify,
+            pkg.manifest_path,
+            features,
+            pkg.config.registry(),
+            args.token.as_ref().map(AsRef::as_ref),
+        )? {
+            return Ok(103);
+        }
+        let timeout = std::time::Duration::from_secs(300);
+
+        if pkg.config.registry().is_none() {
+            let mut index = crates_index::Index::new_cargo_default()?;
+
+            let version = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+            cargo::wait_for_publish(
+                &mut index,
+                crate_name,
+                &version.full_version_string,
+                timeout,
                 dry_run,
-                pkg.config.verify(),
-                pkg.manifest_path,
-                features,
-                pkg.config.registry(),
-                args.token.as_ref().map(AsRef::as_ref),
-            )? {
-                return Ok(103);
-            }
-            let timeout = std::time::Duration::from_secs(300);
-
-            if pkg.config.registry().is_none() {
-                cargo::wait_for_publish(crate_name, &base.full_version_string, timeout, dry_run)?;
-                // HACK: Even once the index is updated, there seems to be another step before the publish is fully ready.
-                // We don't have a way yet to check for that, so waiting for now in hopes everything is ready
-                if !dry_run {
-                    let publish_grace_sleep = std::env::var("PUBLISH_GRACE_SLEEP")
-                        .unwrap_or_else(|_| Default::default())
-                        .parse()
-                        .unwrap_or(0);
-                    if 0 < publish_grace_sleep {
-                        log::info!(
-                            "Waiting an additional {} seconds for crates.io to update its indices...",
-                            publish_grace_sleep
-                        );
-                        std::thread::sleep(std::time::Duration::from_secs(publish_grace_sleep));
-                    }
+            )?;
+            // HACK: Even once the index is updated, there seems to be another step before the publish is fully ready.
+            // We don't have a way yet to check for that, so waiting for now in hopes everything is ready
+            if !dry_run {
+                let publish_grace_sleep = std::env::var("PUBLISH_GRACE_SLEEP")
+                    .unwrap_or_else(|_| Default::default())
+                    .parse()
+                    .unwrap_or(0);
+                if 0 < publish_grace_sleep {
+                    log::info!(
+                        "Waiting an additional {} seconds for crates.io to update its indices...",
+                        publish_grace_sleep
+                    );
+                    std::thread::sleep(std::time::Duration::from_secs(publish_grace_sleep));
                 }
-            } else {
-                log::debug!("Not waiting for publish because the registry is not crates.io and doesn't get updated automatically");
             }
+        } else {
+            log::debug!("Not waiting for publish because the registry is not crates.io and doesn't get updated automatically");
         }
     }
 
     // STEP 5: Tag
+    let mut seen_tags = HashSet::new();
     for pkg in pkgs {
         if let Some(tag_name) = pkg.tag.as_ref() {
-            let sign = pkg.config.sign_commit() || pkg.config.sign_tag();
+            if seen_tags.insert(tag_name) {
+                let cwd = pkg.package_root;
+                let crate_name = pkg.meta.name.as_str();
 
-            // FIXME: remove when the meaning of sign_commit is changed
-            if !pkg.config.sign_tag() && pkg.config.sign_commit() {
-                log::warn!("In next minor release, `sign-commit` will only be used to control git commit signing. Use option `sign-tag` for tag signing.");
-            }
+                let version = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+                let prev_version_var = pkg.prev_version.bare_version_string.as_str();
+                let prev_metadata_var = pkg.prev_version.full_version.build.as_str();
+                let version_var = version.bare_version_string.as_str();
+                let metadata_var = version.full_version.build.as_str();
+                let template = Template {
+                    prev_version: Some(prev_version_var),
+                    prev_metadata: Some(prev_metadata_var),
+                    version: Some(version_var),
+                    metadata: Some(metadata_var),
+                    crate_name: Some(crate_name),
+                    tag_name: Some(tag_name),
+                    date: Some(NOW.as_str()),
+                    ..Default::default()
+                };
+                let tag_message = template.render(pkg.config.tag_message());
 
-            let cwd = pkg.package_path;
-            let crate_name = pkg.meta.name.as_str();
-
-            let base = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
-            let template = Template {
-                prev_version: Some(&pkg.prev_version.full_version_string),
-                version: Some(&base.full_version_string),
-                crate_name: Some(crate_name),
-                tag_name: Some(tag_name),
-                date: Some(NOW.as_str()),
-                ..Default::default()
-            };
-            let tag_message = template.render(pkg.config.tag_message());
-
-            log::debug!("Creating git tag {}", tag_name);
-            if !git::tag(cwd, tag_name, &tag_message, sign, dry_run)? {
-                // tag failed, abort release
-                return Ok(104);
+                log::debug!("Creating git tag {}", tag_name);
+                if !git::tag(cwd, tag_name, &tag_message, pkg.config.sign_tag(), dry_run)? {
+                    // tag failed, abort release
+                    return Ok(104);
+                }
             }
         }
     }
@@ -459,29 +563,40 @@ fn release_packages<'m>(
     let mut shared_commit = false;
     let mut shared_post_version: Option<version::Version> = None;
     for pkg in pkgs {
-        if let Some(version) = pkg.post_version.as_ref() {
-            let cwd = pkg.package_path;
+        if let Some(next_version) = pkg.post_version.as_ref() {
+            let cwd = pkg.package_root;
             let crate_name = pkg.meta.name.as_str();
 
-            let updated_version_string = version.full_version_string.as_ref();
             log::info!(
                 "Starting {}'s next development iteration {}",
                 crate_name,
-                updated_version_string,
+                next_version.full_version_string
             );
-            update_dependent_versions(pkg, version, dry_run)?;
+            update_dependent_versions(pkg, next_version, dry_run)?;
             if !dry_run {
-                cargo::set_package_version(pkg.manifest_path, updated_version_string)?;
+                cargo::set_package_version(
+                    pkg.manifest_path,
+                    next_version.full_version_string.as_str(),
+                )?;
                 cargo::update_lock(pkg.manifest_path)?;
             }
-            let base = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+            let version = pkg.version.as_ref().unwrap_or(&pkg.prev_version);
+            let prev_version_var = pkg.prev_version.bare_version_string.as_str();
+            let prev_metadata_var = pkg.prev_version.full_version.build.as_str();
+            let version_var = version.bare_version_string.as_str();
+            let metadata_var = version.full_version.build.as_str();
+            let next_version_var = next_version.bare_version_string.as_ref();
+            let next_metadata_var = next_version.full_version.build.as_ref();
             let template = Template {
-                prev_version: Some(&pkg.prev_version.full_version_string),
-                version: Some(&base.full_version_string),
+                prev_version: Some(prev_version_var),
+                prev_metadata: Some(prev_metadata_var),
+                version: Some(version_var),
+                metadata: Some(metadata_var),
                 crate_name: Some(crate_name),
                 date: Some(NOW.as_str()),
                 tag_name: pkg.tag.as_deref(),
-                next_version: Some(updated_version_string),
+                next_version: Some(next_version_var),
+                next_metadata: Some(next_metadata_var),
                 ..Default::default()
             };
             if !pkg.config.post_release_replacements().is_empty() {
@@ -496,7 +611,7 @@ fn release_packages<'m>(
             }
 
             if pkg.config.shared_version() && shared_post_version.is_none() {
-                shared_post_version = Some(version.clone());
+                shared_post_version = Some(next_version.clone());
             }
             if pkg.config.consolidate_commits() {
                 shared_commit = true;
@@ -512,14 +627,24 @@ fn release_packages<'m>(
     }
     if shared_commit {
         let shared_commit_msg = {
+            let version_var = shared_version
+                .as_ref()
+                .map(|v| v.bare_version_string.as_str());
+            let metadata_var = shared_version
+                .as_ref()
+                .map(|v| v.full_version.build.as_str());
+            let next_version_var = shared_post_version
+                .as_ref()
+                .map(|v| v.bare_version_string.as_str());
+            let next_metadata_var = shared_post_version
+                .as_ref()
+                .map(|v| v.full_version.build.as_str());
             let template = Template {
-                version: shared_version
-                    .as_ref()
-                    .map(|v| v.full_version_string.as_str()),
+                version: version_var,
+                metadata: metadata_var,
                 date: Some(NOW.as_str()),
-                next_version: shared_post_version
-                    .as_ref()
-                    .map(|v| v.full_version_string.as_str()),
+                next_version: next_version_var,
+                next_metadata: next_metadata_var,
                 ..Default::default()
             };
             template.render(ws_config.post_release_commit_message())
@@ -537,41 +662,37 @@ fn release_packages<'m>(
 
     // STEP 7: git push
     if ws_config.push() {
-        let mut shared_push = false;
+        let mut shared_refs = HashSet::new();
         for pkg in pkgs {
             if !pkg.config.push() {
                 continue;
             }
 
-            let cwd = pkg.package_path;
-            if let Some(tag_name) = pkg.tag.as_ref() {
-                log::info!("Pushing {} to {}", tag_name, git_remote);
-                if !git::push_tag(cwd, git_remote, tag_name, dry_run)? {
-                    return Ok(106);
-                }
-            }
-
             if pkg.config.consolidate_pushes() {
-                shared_push = true
-            } else if !shared_push {
-                log::info!("Pushing HEAD to {}", git_remote);
-                if !git::push(
-                    cwd,
-                    git_remote,
-                    Some(branch.as_str()),
-                    pkg.config.push_options(),
-                    dry_run,
-                )? {
+                shared_refs.insert(branch.as_str());
+                if let Some(tag_name) = pkg.tag.as_deref() {
+                    shared_refs.insert(tag_name);
+                }
+            } else {
+                let mut refs = vec![branch.as_str()];
+                if let Some(tag_name) = pkg.tag.as_deref() {
+                    refs.push(tag_name)
+                }
+                log::info!("Pushing {} to {}", refs.join(", "), git_remote);
+                let cwd = pkg.package_root;
+                if !git::push(cwd, git_remote, refs, pkg.config.push_options(), dry_run)? {
                     return Ok(106);
                 }
             }
         }
-        if shared_push {
-            log::info!("Pushing HEAD to {}", git_remote);
+        if !shared_refs.is_empty() {
+            let mut shared_refs = shared_refs.into_iter().collect::<Vec<_>>();
+            shared_refs.sort_unstable();
+            log::info!("Pushing {} to {}", shared_refs.join(", "), git_remote);
             if !git::push(
                 ws_meta.workspace_root.as_std_path(),
                 git_remote,
-                Some(branch.as_str()),
+                shared_refs,
                 ws_config.push_options(),
                 dry_run,
             )? {
@@ -609,22 +730,21 @@ fn find_dependents<'w>(
 struct PackageRelease<'m> {
     meta: &'m cargo_metadata::Package,
     manifest_path: &'m Path,
-    package_path: &'m Path,
+    package_root: &'m Path,
+    is_root: bool,
     config: config::Config,
 
-    bin: bool,
-
     package_content: Vec<std::path::PathBuf>,
+    bin: bool,
+    dependents: Vec<Dependency<'m>>,
+    features: cargo::Features,
 
     prev_version: version::Version,
     prev_tag: String,
+
     version: Option<version::Version>,
     tag: Option<String>,
     post_version: Option<version::Version>,
-
-    dependents: Vec<Dependency<'m>>,
-
-    features: cargo::Features,
 }
 
 impl<'m> PackageRelease<'m> {
@@ -635,60 +755,40 @@ impl<'m> PackageRelease<'m> {
         pkg_meta: &'m cargo_metadata::Package,
     ) -> Result<Option<Self>, error::FatalError> {
         let manifest_path = pkg_meta.manifest_path.as_std_path();
-        let cwd = manifest_path.parent().unwrap_or_else(|| Path::new("."));
-
-        let config = {
-            let mut release_config = config::Config::default();
-
-            if !args.isolated {
-                let cfg =
-                    config::resolve_config(ws_meta.workspace_root.as_std_path(), manifest_path)?;
-                release_config.update(&cfg);
-            }
-
-            if let Some(custom_config_path) = args.custom_config.as_ref() {
-                // when calling with -c option
-                let cfg = config::resolve_custom_config(Path::new(custom_config_path))?
-                    .unwrap_or_default();
-                release_config.update(&cfg);
-            }
-
-            release_config.update(&args.config.to_config());
-
-            // the publish flag in cargo file
-            let cargo_file = cargo::parse_cargo_config(manifest_path)?;
-            if !cargo_file
-                .get("package")
-                .and_then(|f| f.as_table())
-                .and_then(|f| f.get("publish"))
-                .and_then(|f| f.as_bool())
-                .unwrap_or(true)
-            {
-                release_config.publish = Some(false);
-                release_config.disable_publish = None;
-            }
-
-            release_config
-        };
+        let package_root = manifest_path.parent().unwrap_or_else(|| Path::new("."));
+        let config = config::load_package_config(args, ws_meta, pkg_meta)?;
         if !config.release() {
             log::trace!("Disabled in config, skipping {}", manifest_path.display());
             return Ok(None);
         }
 
-        let is_root = git_root == cwd;
-
-        let prev_version = version::Version::from(pkg_meta.version.clone());
-
         let package_content = cargo::package_content(manifest_path)?;
+        let bin = pkg_meta
+            .targets
+            .iter()
+            .flat_map(|t| t.kind.iter())
+            .any(|k| k == "bin");
+        let features = config.features();
+        let dependents = find_dependents(ws_meta, pkg_meta)
+            .map(|(pkg, dep)| Dependency { pkg, req: &dep.req })
+            .collect();
 
+        let is_root = git_root == package_root;
+        let prev_version = version::Version::from(pkg_meta.version.clone());
         let prev_tag = if let Some(prev_tag) = args.prev_tag_name.as_ref() {
             // Trust the user that the tag passed in is the latest tag for the workspace and that
             // they don't care about any changes from before this tag.
             prev_tag.to_owned()
         } else {
+            let prev_version_var = prev_version.bare_version_string.as_str();
+            let prev_metadata_var = prev_version.full_version.build.as_str();
+            let version_var = prev_version.bare_version_string.as_str();
+            let metadata_var = prev_version.full_version.build.as_str();
             let mut template = Template {
-                prev_version: Some(&prev_version.full_version_string),
-                version: Some(&prev_version.full_version_string),
+                prev_version: Some(prev_version_var),
+                prev_metadata: Some(prev_metadata_var),
+                version: Some(version_var),
+                metadata: Some(metadata_var),
                 crate_name: Some(pkg_meta.name.as_str()),
                 ..Default::default()
             };
@@ -702,70 +802,70 @@ impl<'m> PackageRelease<'m> {
         let version = args
             .level_or_version
             .bump(&prev_version.full_version, args.metadata.as_deref())?;
-        let is_pre_release = version
-            .as_ref()
-            .map(version::Version::is_prerelease)
-            .unwrap_or(false);
-        let dependents = find_dependents(ws_meta, pkg_meta)
-            .map(|(pkg, dep)| Dependency { pkg, req: &dep.req })
-            .collect();
+        let tag = None;
+        let post_version = None;
 
-        let base = version.as_ref().unwrap_or(&prev_version);
+        let pkg = PackageRelease {
+            meta: pkg_meta,
+            manifest_path,
+            package_root,
+            is_root,
+            config,
 
-        let tag = if config.tag() {
+            package_content,
+            bin,
+            dependents,
+            features,
+
+            prev_version,
+            prev_tag,
+
+            version,
+            tag,
+            post_version,
+        };
+        Ok(Some(pkg))
+    }
+
+    fn plan(&mut self) -> Result<(), FatalError> {
+        let base = self.version.as_ref().unwrap_or(&self.prev_version);
+        let tag = if self.config.tag() {
+            let prev_version_var = self.prev_version.bare_version_string.as_str();
+            let prev_metadata_var = self.prev_version.full_version.build.as_str();
+            let version_var = base.bare_version_string.as_str();
+            let metadata_var = base.full_version.build.as_str();
             let mut template = Template {
-                prev_version: Some(&prev_version.full_version_string),
-                version: Some(&base.full_version_string),
-                crate_name: Some(pkg_meta.name.as_str()),
+                prev_version: Some(prev_version_var),
+                prev_metadata: Some(prev_metadata_var),
+                version: Some(version_var),
+                metadata: Some(metadata_var),
+                crate_name: Some(self.meta.name.as_str()),
                 ..Default::default()
             };
 
-            let tag_prefix = config.tag_prefix(is_root);
+            let tag_prefix = self.config.tag_prefix(self.is_root);
             let tag_prefix = template.render(tag_prefix);
             template.prefix = Some(&tag_prefix);
-            Some(template.render(config.tag_name()))
+            Some(template.render(self.config.tag_name()))
         } else {
             None
         };
 
-        let post_version = if !is_pre_release && config.dev_version() {
+        let is_pre_release = base.is_prerelease();
+        let post_version = if !is_pre_release && self.config.dev_version() {
             let mut post = base.full_version.clone();
             post.increment_patch();
-            post.pre = semver::Prerelease::new(config.dev_version_ext())?;
+            post.pre = semver::Prerelease::new(self.config.dev_version_ext())?;
 
             Some(version::Version::from(post))
         } else {
             None
         };
 
-        let bin = pkg_meta
-            .targets
-            .iter()
-            .flat_map(|t| t.kind.iter())
-            .any(|k| k == "bin");
+        self.tag = tag;
+        self.post_version = post_version;
 
-        let features = config.features();
-
-        let pkg = PackageRelease {
-            meta: pkg_meta,
-            manifest_path,
-            package_path: cwd,
-            config,
-
-            bin,
-
-            package_content,
-
-            prev_version,
-            prev_tag,
-            version,
-            tag,
-            post_version,
-            dependents,
-
-            features,
-        };
-        Ok(Some(pkg))
+        Ok(())
     }
 }
 
@@ -779,7 +879,7 @@ fn changed_since<'m>(
         ws_meta.workspace_root.as_std_path()
     } else {
         // Limit our lookup since we don't need to check for `Cargo.lock`
-        pkg.package_path
+        pkg.package_root
     };
     let changed = git::changed_files(changed_root, since_ref).ok().flatten()?;
     let mut changed: Vec<_> = changed
